data Op  = Add | Sub | Mul | Div
 deriving (Eq,Show)
data Expr = Val Int | App Op Expr Expr
 deriving (Eq,Show)

oplist :: [Op]
oplist = [Add,Sub,Mul,Div]

subs :: [a] -> [[a]]

subs [] = [[]]
subs (x:xs) =  (subs xs) ++ map (x:) (subs xs)

inter x [] = [[x]]
inter x (y:ys) = (x:y:ys):(map (y:) (inter x ys))

perm [] = [[]]
perm (x:xs) =  concatMap (inter x) (perm xs)

choices :: [a] -> [[a]]
choices l = concatMap perm (subs l)

--- returns empty list or singleton
validexpns :: Op ->  (Int,Expr) -> (Int,Expr) -> [(Int,Expr)]
validexpns op (v1,e1) (v2,e2) 
	| (op == Add) = [(v1 + v2, App Add e1 e2)]
	| (op == Mul) = [(v1 * v2,App Mul e1 e2)]
	| (op == Sub) = if (v2 <= v1) then [(v1 - v2, App Sub e1 e2)] else []
	| (op == Div) = if ((v2/=0)&&(mod v1 v2 ==0)) then [(div v1 v2, App Div e1 e2)] else []

genandeval :: [Int] -> [(Int,Expr)]
genandeval [] = []
genandeval [x] = [(x,Val x)]
genandeval l   = concat [ validexpns op x1 x2  | op <- oplist, i <- [1..(length l)-1], 
                             x1 <- genandeval (take i l), x2 <- genandeval (drop i l) ]

solve ::  [Int] -> Int -> String
solve l v = printexp (snd (head (filter (f v) (concatMap genandeval (choices l)))))

f v (x,y)  =  (x == v)

printexp :: Expr -> String
printexp (Val x) = show x
printexp (App Add x y) = "(" ++ (printexp x) ++ "+" ++ (printexp y) ++")" 
printexp (App Sub x y) = "(" ++ (printexp x) ++ "-" ++ (printexp y) ++")" 
printexp (App Mul x y) = "(" ++ (printexp x) ++ "*" ++ (printexp y) ++")" 
printexp (App Div x y) = "(" ++ (printexp x) ++ "/" ++ (printexp y) ++")" 
